"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const printer = ts.createPrinter({
    newLine: ts.NewLineKind.LineFeed,
});
function toSourceCode(node, source) {
    return printer.printNode(ts.EmitHint.Unspecified, node, source || node.getSourceFile());
}
exports.toSourceCode = toSourceCode;
function printAllChildren(node, depth = 0) {
    console.log(new Array(depth + 1).join("----"), ts.formatSyntaxKind(node.kind), node.pos, node.end);
    depth++;
    node.forEachChild(c => printAllChildren(c, depth));
}
exports.printAllChildren = printAllChildren;
function referencesIdentifier(node, name) {
    let seen = false;
    function references(node) {
        if (seen)
            return;
        if (ts.isIdentifier(node)) {
            if (node.escapedText.toString() === name) {
                seen = true;
                return;
            }
        }
        ts.forEachChild(node, references);
    }
    references(node);
    return seen;
}
exports.referencesIdentifier = referencesIdentifier;
// at every potential repeat point, we inject a check if we might be out of budget
function createInstrument() {
    const fn = ts.createElementAccess(ts.createIdentifier("window"), ts.createLiteral("__checkBudget__"));
    return ts.createStatement(ts.createCall(fn, [], []));
}
function instrumentIterator(node) {
    const block = node.statement;
    if (ts.isBlock(block)) {
        return ts.updateBlock(block, [createInstrument()].concat(block.statements));
    }
    else {
        return ts.createBlock([createInstrument(), block]);
    }
}
function instrumentVisitor(node) {
    node = ts.visitEachChild(node, instrumentVisitor, _ctx);
    if (ts.isWhileStatement(node)) {
        return ts.updateWhile(node, node.expression, instrumentIterator(node));
    }
    if (ts.isForStatement(node)) {
        return ts.updateFor(node, node.initializer, node.condition, node.incrementor, instrumentIterator(node));
    }
    if (ts.isFunctionDeclaration(node) && node.body) {
        const body = node.body;
        return ts.updateFunctionDeclaration(node, node.decorators, node.modifiers, node.asteriskToken, node.name, node.typeParameters, node.parameters, node.type, ts.updateBlock(body, [createInstrument()].concat(body.statements)));
    }
    if (ts.isFunctionExpression(node)) {
        const body = node.body;
        return ts.updateFunctionExpression(node, node.modifiers, node.asteriskToken, node.name, node.typeParameters, node.parameters, node.type, ts.updateBlock(body, [createInstrument()].concat(body.statements)));
    }
    if (ts.isMethodDeclaration(node) && node.body) {
        const body = node.body;
        return ts.updateMethod(node, node.decorators, node.modifiers, node.asteriskToken, node.name, node.questionToken, node.typeParameters, node.parameters, node.type, ts.updateBlock(body, [createInstrument()].concat(body.statements)));
    }
    if (ts.isArrowFunction(node) && ts.isBlock(node.body)) {
        const body = node.body;
        return ts.updateArrowFunction(node, node.modifiers, node.typeParameters, node.parameters, node.type, ts.updateBlock(body, [createInstrument()].concat(body.statements)));
    }
    if (ts.isConstructorDeclaration(node) && node.body) {
        const body = node.body;
        return ts.updateConstructor(node, node.decorators, node.modifiers, node.parameters, ts.updateBlock(body, [createInstrument()].concat(body.statements)));
    }
    return node;
}
exports.instrumentVisitor = instrumentVisitor;
var ComponentType;
(function (ComponentType) {
    ComponentType["Component"] = "component";
    ComponentType["Override"] = "override";
    // Device and DeviceSkin are currently not used in code, only in package.json of Framer Library, keeping for completeness
    ComponentType["Device"] = "device";
    ComponentType["DeviceSkin"] = "deviceSkin";
})(ComponentType = exports.ComponentType || (exports.ComponentType = {}));
function buildInfoExport(components) {
    const dicts = components.map(c => {
        const properties = [ts.createPropertyAssignment("name", ts.createLiteral(c.name))];
        if (c.children !== undefined) {
            properties.push(ts.createPropertyAssignment("children", ts.createLiteral(c.children)));
        }
        properties.push(ts.createPropertyAssignment("type", ts.createLiteral(c.type)));
        return ts.createObjectLiteral(properties);
    });
    const info = ts.createVariableDeclaration("__info__", undefined, ts.createArrayLiteral(dicts));
    return ts.createVariableStatement([ts.createToken(ts.SyntaxKind.ExportKeyword)], [info]);
}
exports.buildInfoExport = buildInfoExport;
function analyzeClasses(node, classes, exports) {
    if (!ts.isClassDeclaration(node))
        return;
    // figure out name
    let name = "Default";
    if (node.name) {
        name = node.name.getText();
    }
    if (node.modifiers) {
        for (let x of node.modifiers) {
            if (x.kind === ts.SyntaxKind.ExportKeyword) {
                exports.add(name);
                break;
            }
        }
    }
    // must extend Component, we blindly assume that means React.Component
    let extendsComponent = false;
    if (node.heritageClauses) {
        for (let x of node.heritageClauses) {
            if (referencesIdentifier(x, "Component")) {
                extendsComponent = true;
                break;
            }
        }
    }
    if (!extendsComponent)
        return;
    // figure out if this node references this.props.children
    // there are so many ways it can, we merely look for a `children` identifier being used
    const children = referencesIdentifier(node, "children");
    const type = ComponentType.Component;
    classes.set(name, { children, type });
}
function analyzeHigherOrderComponent(node, name, classes) {
    if (!node)
        return;
    if (!ts.isCallExpression(node))
        return;
    node.arguments.forEach(n => {
        if (ts.isIdentifier(n)) {
            const ref = n.escapedText.toString();
            if (classes.has(ref)) {
                // a call expression that direcly references a name that refers to a known component class
                classes.set(name, classes.get(ref));
            }
        }
    });
}
function analyzeOverride(node, name, classes) {
    if (!node)
        return;
    if (!node.initializer)
        return;
    if (!ts.isFunctionLike(node.initializer))
        return;
    const type = node.type;
    if (type &&
        ts.isTypeReferenceNode(type) &&
        ts.isIdentifier(type.typeName) &&
        (type.typeName.escapedText === "Override" || type.typeName.escapedText === "OverrideFunction")) {
        classes.set(name, { type: ComponentType.Override });
    }
}
function analyzeExportedVariableDeclaration(node, classes, exports, exported) {
    // not all nodes seem to be connected to a source file in practice
    if (!node.getSourceFile())
        return;
    node.declarations.forEach(n => {
        if (ts.isVariableDeclaration(n)) {
            const name = n.name.getText();
            if (exported)
                exports.add(name);
            analyzeHigherOrderComponent(n.initializer, name, classes);
            analyzeOverride(n, name, classes);
        }
    });
}
function analyzeExportedIdentifiers(node, exports) {
    if (ts.isIdentifier(node)) {
        exports.add(node.escapedText.toString());
    }
    ts.forEachChild(node, n => analyzeExportedIdentifiers(n, exports));
}
function analyzeExports(node, classes, exports) {
    if (ts.isClassDeclaration(node))
        return;
    if (ts.isVariableStatement(node)) {
        let exported = false;
        if (node.modifiers) {
            for (let x of node.modifiers) {
                if (x.kind === ts.SyntaxKind.ExportKeyword) {
                    exported = true;
                    break;
                }
            }
        }
        // also find higher order components if we find `ident = call(..., ident, ...)`
        analyzeExportedVariableDeclaration(node.declarationList, classes, exports, exported);
        return;
    }
    if (ts.isExportDeclaration(node)) {
        analyzeExportedIdentifiers(node, exports);
        return;
    }
    if (ts.isExportAssignment(node)) {
        analyzeExportedIdentifiers(node, exports);
        return;
    }
    if (ts.isExportSpecifier(node)) {
        analyzeExportedIdentifiers(node, exports);
        return;
    }
}
function analyzeFile(sourceFile) {
    const statements = [];
    const exportedNames = new Set();
    const componentClasses = new Map();
    ts.forEachChild(sourceFile, (n) => {
        if (ts.isVariableStatement(n)) {
            if (referencesIdentifier(n, "___info__") || referencesIdentifier(n, "__info__")) {
                return;
            }
        }
        statements.push(instrumentVisitor(n));
        analyzeClasses(n, componentClasses, exportedNames);
        analyzeExports(n, componentClasses, exportedNames);
    });
    if (exportedNames.size > 0) {
        const results = [];
        for (let name of componentClasses.keys()) {
            if (!exportedNames.has(name))
                continue;
            const { children, type } = componentClasses.get(name);
            results.push({ name, children, type });
        }
        statements.push(buildInfoExport(results));
    }
    return ts.updateSourceFileNode(sourceFile, ts.createNodeArray(statements));
}
let _ctx;
function analyzeAndExportComponents(ctx) {
    return (sourceFile) => {
        _ctx = ctx;
        return analyzeFile(sourceFile);
    };
}
exports.analyzeAndExportComponents = analyzeAndExportComponents;
